//***********************************************************
//
//  NXV4 "Math-on-Paper" software
//
//  package : CPX class _____________________________________
//            
//            supports: move semantix
//                      operators + - * / ^   L,R-val context
//                                += -= *= /=
//                      fabs()
//                      phi ()    [ 0  ... 2.pi] o/p
//                      phi2()    [-pi ...  +pi] o/p
//                      ~  -      unary operators L,R-val
//                      cpx<arithmetic> i.e. int
//                                           float 
//                                           double
//                                           long double
//                      type_casts between the types 
//                      cout << overload 
//
//                      TOTAL = 27+1205 instantiated op/f's
//                      _____________________________________
//
//
//  authors : Mihai-Tiberiu Dima (1)
//            Maria Dima         (2)
//
//             Meshcheryakov Laboratory of Information Techn.
//             Joint Institute for Nuclear Research
//             Dubna-Russia
//
//            on leave from:
//
//             Hyperion Univ. Bucharest-Romania
//
//
//  date    : CPX4/v1.0.alfa   / Mon Dec 27 17:13:57 CET 2021
//                 v2.1.beta   / Mon Mar 07 12:45:54 CET 2022
//
//***********************************************************

#pragma once


#include<iostream>
#include<cmath>
#include<type_traits>
#include<utility>

using std::cout                                             ;
using std::endl                                             ;
using std::ostream                                          ;
using std::decay                                            ;
using std::enable_if                                        ;
using std::is_same                                          ;
using std::is_arithmetic                                    ;
using std::declval                                          ;

namespace cpx4
{
 using fltx =      float                                    ;
 using dblx =      double                                   ;
 using real = long double                                   ;

 long double PI  = 3.14159265359                            ;
 long double lim = 1.0e-160                                 ;
 long double lm2 = std::sqrt(lim)                           ;
 long double LIM = 1/lim                                    ;
 long double LD2 = std::sqrt(LIM)                           ;
 long double AD2 = std::log(LD2)                            ;


 template<typename scalar=double>
 class cpx
   {public: typename decay<scalar>::type x = 0              ;
            typename decay<scalar>::type y = 0              ;
      // 
      cpx()                   // plain     
       {                                                    ;} 
      //
      cpx(     scalar a )     // manual     
       {x = a                                               ;
        y = 0                                               ;} 
      //
      cpx(     scalar a,
               scalar b )     // manual     
       {x = a                                               ;
        y = b                                               ;} 
      //
      cpx(const cpx& A)       // copy       
       {x = A.x                                             ;
        y = A.y                                             ;} 

      template<typename U>
      cpx(cpx<U> const& A)
       {x = static_cast<scalar>(A.x)                        ;
        y = static_cast<scalar>(A.y)                        ;}

      // CAST  _____________________________________________

      explicit
      operator int ()        {return    int (x)             ;}

      explicit
      operator fltx()        {return    fltx(x)             ;}

      explicit
      operator dblx()        {return    dblx(x)             ;}

      explicit
      operator real()        {return    real(x)             ;}

      explicit
      operator cpx<int >()   {return cpx<int >(int (x), int (y));}

      explicit
      operator cpx<fltx>()   {return cpx<fltx>(fltx(x), fltx(y));}

      explicit
      operator cpx<dblx>()   {return cpx<dblx>(dblx(x), dblx(y));}

      explicit
      operator cpx<real>()   {return cpx<real>(real(x), real(y));}

      // ASSGN _____________________________________________

      scalar& operator= (const scalar& A)
       {x = A                                               ;
        y = 0                                               ;
        return const_cast<scalar&>(A)                       ;}

         cpx& operator= (const    cpx& A)
       {x = A.x                                             ;
        y = A.y                                             ;
        return const_cast<cpx&>(A)                          ;}

      template<typename U> 
      cpx<U>& operator= (cpx<U> const& A)
       {x = static_cast<scalar>(A.x)                        ;
        y = static_cast<scalar>(A.y)                        ;
        return const_cast<cpx<U>&>(A)                       ;}

      // NGTV ______________________________________________

      template<typename T>
      friend
      cpx<T>
      operator- (cpx<T> const& A)                           ;

      template<typename T>
      friend
      cpx<T>&&
      operator- (cpx<T> && A)                               ;

      // TILDE (cpx.conj.) _________________________________

      template<typename T>
      friend
      cpx<T>
      operator~ (cpx<T> const& A)                           ;

      template<typename T>
      friend
      cpx<T>&&
      operator~ (cpx<T> && A)                               ;

      // FABS ______________________________________________

      template<typename T>
      friend
      T fabs(cpx<T> const& A)                               ;

      template<typename T>
      friend
      T n2(cpx<T> const& A)                                 ;

      // PHI _______________________________________________

      template<typename T>           //  o/p =  [0 ... +2PI]
      friend
      T phi (const cpx<T>& A)                               ;

      // PHI2 ______________________________________________

      template<typename T>           //  o/p = [-PI ... +PI] 
      friend
      T phi2(const cpx<T>& A)                               ;

      // PROD ______________________________________________

      template<typename L, typename R>
      friend
      cpx<typename decay<decltype(declval<L>()*
                                  declval<R>())>::type>
      operator|(cpx<L> const& A, cpx<R> const& B)           ;
      //
      //
      //
      #include "comp.cxx"
      #include "eqal.cxx"

      #include "plus.cxx"
      #include "mnus.cxx"
      #include "mult.cxx"
      #include "divd.cxx"

      #include "expo.cxx"
      #include "alog.cxx"
      //
      // ....................................................
      // 
      friend std::ostream& operator<<(std::ostream&   , 
                                      cpx<auto> const& )    ;
      //
      //-----------------------------------------------------
      //
     ~cpx()
       {                                                    ;}
 };

//===========================================================
};
